#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#
"""Paths utils."""

from pathlib import Path

def is_relative_to(path: Path, *other):
	"""Return True if the path is relative to another path or False.
	"""
	try:
		path.relative_to(*other)
		return True
	except ValueError:
		return False



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#

from functools import cmp_to_key
from locale import strcoll
from pathlib import Path
from typing import List
from sebaubuntu_libs.liblogging import LOGW
from xml.etree import ElementTree
from textwrap import indent

from sebaubuntu_libs.libandroid.vintf import INDENTATION
from sebaubuntu_libs.libandroid.vintf.aidl import AidlHal
from sebaubuntu_libs.libandroid.vintf.common import Hal
from sebaubuntu_libs.libandroid.vintf.hidl import HidlHal

def strcoll_hal(obj1: Hal, obj2: Hal) -> int:
	# Sort by name if different
	if obj1.name != obj2.name:
		return strcoll(obj1.name, obj2.name)

	# AIDL first
	if isinstance(obj1, AidlHal) and not isinstance(obj2, AidlHal):
		return -1
	if isinstance(obj2, AidlHal) and not isinstance(obj1, AidlHal):
		return 1

	# transport=hwbinder first if both HIDL
	if isinstance(obj1, HidlHal) and isinstance(obj2, HidlHal) and obj1.transport != obj2.transport:
		return 1 if obj1.transport == "hwbinder" else -1

	# Compare normally
	return strcoll(obj1.name, obj2.name)

class Manifest:
	"""A class representing a VINTF manifest."""
	def __init__(self):
		"""Parse a VINTF manifest."""
		self.version = None
		self.type: str = None
		self.target_level: str = None
		self.entries: List[Hal] = []

	def __str__(self):
		string = f'<manifest version="{self.version}" type="{self.type}" target-level="{self.target_level}">\n'
		for entry in sorted(self.entries, key=cmp_to_key(strcoll_hal)):
			string += indent(f"{entry}\n", INDENTATION)
		string += "</manifest>\n"

		return string

	def import_file(self, file: Path):
		"""Import a manifest file."""
		tree = ElementTree.parse(file)
		root = tree.getroot()

		if self.version is None and "version" in root.attrib:
			self.version = root.attrib["version"]
		if self.type is None and "type" in root.attrib:
			self.type = root.attrib["type"]
		if self.target_level is None and "target-level" in root.attrib:
			self.target_level = root.attrib["target-level"]

		# Parse HALs
		for entry in root:
			if entry.tag != "hal":
				continue

			hal_format = entry.get("format")
			if hal_format == "aidl":
				self.entries.append(AidlHal.from_entry(entry))
			elif hal_format == "hidl":
				self.entries.append(HidlHal.from_entry(entry))
			else:
				LOGW(f"Unknown HAL type {hal_format}")



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#
"""AIK wrapper library."""

from git import Repo
from pathlib import Path
from platform import system
from sebaubuntu_libs.liblogging import LOGD, LOGI
from shutil import which
from subprocess import check_output, STDOUT, CalledProcessError
from tempfile import TemporaryDirectory

AIK_REPO = "https://github.com/SebaUbuntu/AIK-Linux-mirror"

ALLOWED_OS = [
	"Linux",
	"Darwin",
]

class AIKImageInfo:
	def __init__(self,
	             base_address: str,
	             board_name: str,
	             cmdline: str,
	             dt: Path,
	             dtb: Path,
	             dtb_offset: str,
	             dtbo: Path,
	             header_version: str,
	             image_type: str,
	             kernel: Path,
	             kernel_offset: str,
	             origsize: str,
	             os_version: str,
	             pagesize: str,
	             ramdisk: Path,
	             ramdisk_compression: str,
	             ramdisk_offset: str,
	             sigtype: str,
	             tags_offset: str,
	            ):
		self.kernel = kernel
		self.dt = dt
		self.dtb = dtb
		self.dtbo = dtbo
		self.ramdisk = ramdisk
		self.base_address = base_address
		self.board_name = board_name
		self.cmdline = cmdline
		self.dtb_offset = dtb_offset
		self.header_version = header_version
		self.image_type = image_type
		self.kernel_offset = kernel_offset
		self.origsize = origsize
		self.os_version = os_version
		self.pagesize = pagesize
		self.ramdisk_compression = ramdisk_compression
		self.ramdisk_offset = ramdisk_offset
		self.sigtype = sigtype
		self.tags_offset = tags_offset

	def __str__(self):
		return (
			f"base address: {self.base_address}\n"
			f"board name: {self.board_name}\n"
			f"cmdline: {self.cmdline}\n"
			f"dtb offset: {self.dtb_offset}\n"
			f"header version: {self.header_version}\n"
			f"image type: {self.image_type}\n"
			f"kernel offset: {self.kernel_offset}\n"
			f"original size: {self.origsize}\n"
			f"os version: {self.os_version}\n"
			f"page size: {self.pagesize}\n"
			f"ramdisk compression: {self.ramdisk_compression}\n"
			f"ramdisk offset: {self.ramdisk_offset}\n"
			f"sigtype: {self.sigtype}\n"
			f"tags offset: {self.tags_offset}\n"
		)

class AIKManager:
	"""
	This class is responsible for dealing with AIK tasks
	such as cloning, updating, and extracting recovery images.
	"""

	def __init__(self):
		"""Initialize AIKManager class."""
		if system() not in ALLOWED_OS:
			raise NotImplementedError(f"{system()} is not supported")

		# Check whether cpio package is installed
		if which("cpio") is None:
			raise RuntimeError("cpio package is not installed")

		self.tempdir = TemporaryDirectory()
		self.path = Path(self.tempdir.name)

		self.images_path = self.path / "split_img"
		self.ramdisk_path = self.path / "ramdisk"

		LOGI("Cloning AIK...")
		Repo.clone_from(AIK_REPO, self.path)

	def unpackimg(self, image: Path):
		"""Extract recovery image."""
		image_prefix = image.name

		try:
			process = self._execute_script("unpackimg.sh", image)
		except CalledProcessError as e:
			returncode = e.returncode
			output = e.output
		else:
			returncode = 0
			output = process

		if returncode != 0:
			LOGD(output)
			raise RuntimeError(f"AIK extraction failed, return code {returncode}")

		return self.get_current_extracted_info(image_prefix)

	def repackimg(self):
		return self._execute_script("repack.sh")

	def cleanup(self):
		return self._execute_script("cleanup.sh")

	def get_current_extracted_info(self, prefix: str):
		cmdline = None
		for name in ["cmdline", "vendor_cmdline"]:
			_cmdline = self._read_recovery_file(prefix, name)
			if _cmdline:
				cmdline = _cmdline

		dt = self._get_extracted_info(prefix, "dt")
		dt = dt if dt.is_file() else None

		dtb = self._get_extracted_info(prefix, "dtb")
		dtb = dtb if dtb.is_file() else None

		dtbo = None
		for name in ["dtbo", "recovery_dtbo"]:
			_dtbo = self._get_extracted_info(prefix, name)
			if _dtbo.is_file():
				dtbo = _dtbo

		kernel = self._get_extracted_info(prefix, "kernel")
		kernel = kernel if kernel.is_file() else None

		ramdisk = self.ramdisk_path if self.ramdisk_path.is_dir() else None

		ramdisk_compression = None
		for name in ["ramdiskcomp", "vendor_ramdiskcomp"]:
			_ramdisk_compression = self._read_recovery_file(prefix, name)
			if ramdisk_compression:
				ramdisk_compression = _ramdisk_compression

		return AIKImageInfo(
			base_address=self._read_recovery_file(prefix, "base"),
			board_name=self._read_recovery_file(prefix, "board"),
			cmdline=cmdline,
			dt=dt,
			dtb=dtb,
			dtb_offset=self._read_recovery_file(prefix, "dtb_offset"),
			dtbo=dtbo,
			header_version=self._read_recovery_file(prefix, "header_version", default="0"),
			image_type=self._read_recovery_file(prefix, "imgtype"),
			kernel=kernel,
			kernel_offset=self._read_recovery_file(prefix, "kernel_offset"),
			origsize=self._read_recovery_file(prefix, "origsize"),
			os_version=self._read_recovery_file(prefix, "os_version"),
			pagesize=self._read_recovery_file(prefix, "pagesize"),
			ramdisk=ramdisk,
			ramdisk_compression=ramdisk_compression,
			ramdisk_offset=self._read_recovery_file(prefix, "ramdisk_offset"),
			sigtype=self._read_recovery_file(prefix, "sigtype"),
			tags_offset=self._read_recovery_file(prefix, "tags_offset"),
		)

	def _read_recovery_file(self, prefix: str, fragment: str, default: str = None) -> str:
		file = self._get_extracted_info(prefix, fragment)
		return file.read_text().splitlines()[0].strip() if file.exists() else default

	def _get_extracted_info(self, prefix: str, fragment: str) -> Path:
		return self.images_path / f"{prefix}-{fragment}"

	def _execute_script(self, script: str, *args):
		command = [self.path / script, "--nosudo", *args]
		return check_output(command, stderr=STDOUT, universal_newlines=True, encoding="utf-8")



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#

from textwrap import indent
from typing import Set
from xml.etree.ElementTree import Element

from sebaubuntu_libs.libandroid.vintf import INDENTATION
from sebaubuntu_libs.libandroid.vintf.common import Hal, cast_to_str_key

class AidlInterface:
	"""Class representing a AIDL HAL."""
	def __init__(self, name: str, instance: str):
		"""Initialize an object."""
		self.name = name
		self.instance = instance

	def __str__(self) -> str:
		return f"{self.name}/{self.instance}"

	def __eq__(self, __o: object) -> bool:
		if isinstance(__o, AidlInterface):
			return (self.name == __o.name
			        and self.instance == __o.instance)
		return False

	def __hash__(self) -> int:
		return hash((self.name, self.instance))

	@classmethod
	def from_fqname(cls, string: str) -> 'AidlInterface':
		"""
		Create a AIDL HAL from a FQName.

		Example:
		ILights/default
		"""
		name, instance = string.split("/", 1)

		return cls(name, instance)

class AidlHal(Hal):
	"""Class representing a AIDL HAL."""
	def __init__(self, name: str, interfaces: Set[AidlInterface]):
		"""Initialize an object."""
		super().__init__(name)

		self.interfaces = interfaces

	def __str__(self) -> str:
		string = f'<hal format="aidl">\n'
		string += indent(f'<name>{self.name}</name>\n', INDENTATION)
		for interface in sorted(self.interfaces, key=cast_to_str_key):
			string += indent(f'<fqname>{interface}</fqname>\n', INDENTATION)
		string += '</hal>'

		return string

	@classmethod
	def from_entry(cls, entry: Element) -> 'AidlHal':
		"""Create a AIDL HAL from a VINTF entry."""
		assert entry.get("format") == "aidl"

		name = entry.findtext("name")
		interfaces = set([AidlInterface.from_fqname(interface.text)
		                  for interface in entry.findall("fqname")])

		return cls(name, interfaces)



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#

PARTITIONS = [
	"bootimage",
	"odm",
	"odm_dlkm",
	"product",
	"system",
	"system_ext",
	"vendor",
	"vendor_dlkm",
]

def get_partition_props(format_string: str, add_empty: bool = False):
	"""
	Get a list of props given a string to format.

	If add_empty is True, you need to omit a dot at the end of the partition
	(e.g. "ro.{}.build.date" if add_empty is False, "ro.{}build.date" otherwise).
	"""
	partitions_formatted = [
		(f"{part}." if add_empty else part) for part in PARTITIONS
	]

	partition_props = [format_string.format(partition) for partition in partitions_formatted]
	if add_empty:
		partition_props.append(format_string.format(""))

	return partition_props

def fingerprint_to_description(fingerprint: str):
	_, temp = fingerprint.split("/", 1) # brand
	product, temp = temp.split("/", 1)
	_, temp = temp.split(":", 1) # device
	platform_version, temp = temp.split("/", 1)
	build_id, temp = temp.split("/", 1)
	build_number, temp = temp.split(":", 1)
	build_variant, build_version_tags = temp.split("/", 1)

	return f"{product}-{build_variant} {platform_version} {build_id} {build_number} {build_version_tags}"



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#

from pathlib import Path
from typing import Dict

from sebaubuntu_libs.libandroid.partitions.partition import AndroidPartition, BUILD_PROP_LOCATION
from sebaubuntu_libs.libandroid.partitions.partition_model import SSI, TREBLE, PartitionModel

class Partitions:
	def __init__(self, dump_path: Path):
		self.dump_path = dump_path

		self.partitions: Dict[PartitionModel, AndroidPartition] = {}

		# Search for system
		for system in [self.dump_path / "system", self.dump_path / "system/system"]:
			for build_prop_location in BUILD_PROP_LOCATION:
				if not (system / build_prop_location).is_file():
					continue

				self.partitions[PartitionModel.SYSTEM] = AndroidPartition(PartitionModel.SYSTEM, system)

		assert PartitionModel.SYSTEM in self.partitions
		self.system = self.partitions[PartitionModel.SYSTEM]

		# Search for vendor
		for vendor in [self.partitions[PartitionModel.SYSTEM].path / "vendor", self.dump_path / "vendor"]:
			for build_prop_location in BUILD_PROP_LOCATION:
				if not (vendor / build_prop_location).is_file():
					continue

				self.partitions[PartitionModel.VENDOR] = AndroidPartition(PartitionModel.VENDOR, vendor)

		assert PartitionModel.VENDOR in self.partitions
		self.vendor = self.partitions[PartitionModel.VENDOR]

		# Search for the other partitions
		for model in [model for model in PartitionModel.from_group(SSI) if not (model is PartitionModel.SYSTEM)]:
			self._search_for_partition(model)

		for model in [model for model in PartitionModel.from_group(TREBLE) if not (model is PartitionModel.VENDOR)]:
			self._search_for_partition(model)

	def get_partition(self, model: PartitionModel):
		if not model:
			return None

		if model in self.partitions:
			return self.partitions[model]

		return None

	def get_partition_by_name(self, name: str):
		return self.get_partition(PartitionModel.from_name(name))

	def get_all_partitions(self):
		return self.partitions.values()

	def _search_for_partition(self, model: PartitionModel):
		possible_locations = [
			self.partitions[PartitionModel.SYSTEM].path / model.name,
			self.partitions[PartitionModel.VENDOR].path / model.name,
			self.dump_path / model.name
		]

		for location in possible_locations:
			for build_prop_location in BUILD_PROP_LOCATION:
				if not (location / build_prop_location).is_file():
					continue

				self.partitions[model] = AndroidPartition(model, location)



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#
"""Typing utils."""

from typing import Iterable

def is_iterable_and_not_str(obj: object):
    """Check if an object is iterable and not a string."""
    return isinstance(obj, Iterable) and not isinstance(obj, str)



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#
"""Strings utils."""

def removeprefix(string: str, prefix: str) -> str:
	"""Remove a prefix from a string."""
	if string.startswith(prefix):
		return string[len(prefix):]
	return string

def removesuffix(string: str, suffix: str) -> str:
	"""Remove a suffix from a string."""
	if string.endswith(suffix):
		return string[:-len(suffix)]
	return string



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: Apache-2.0
#
"""Logging utils."""

from logging import basicConfig, INFO, DEBUG
from logging import debug, info, warning, error, fatal

LOGD = debug
LOGI = info
LOGW = warning
LOGE = error
LOGF = fatal

def setup_logging(verbose: bool = False):
	if verbose:
		basicConfig(format='[%(filename)s:%(lineno)s %(levelname)s] %(funcName)s: %(message)s',
		            level=DEBUG)
	else:
		basicConfig(format='[%(levelname)s] %(message)s', level=INFO)
